# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: requests.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'requests.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from . import task_pb2 as task__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0erequests.proto\x12\x17hyrex.performanceserver\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\ntask.proto\"\x94\x04\n\x0e\x45nqueueRequest\x12\n\n\x02id\x18\x01 \x01(\t\x12\x12\n\ndurable_id\x18\x02 \x01(\t\x12\x0f\n\x07root_id\x18\x03 \x01(\t\x12\x1c\n\x0fworkflow_run_id\x18\x04 \x01(\tH\x00\x88\x01\x01\x12\x1d\n\x15workflow_dependencies\x18\x05 \x03(\t\x12\x16\n\tparent_id\x18\x06 \x01(\tH\x01\x88\x01\x01\x12\x11\n\ttask_name\x18\x07 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x08 \x01(\x0c\x12\r\n\x05queue\x18\t \x01(\t\x12\x13\n\x0bmax_retries\x18\n \x01(\x05\x12\x1b\n\x0e\x61ttempt_number\x18\x0b \x01(\x05H\x02\x88\x01\x01\x12\x33\n\x08priority\x18\x0c \x01(\x0e\x32!.hyrex.performanceserver.Priority\x12\x1c\n\x0ftimeout_seconds\x18\r \x01(\x05H\x03\x88\x01\x01\x12\x1c\n\x0fidempotency_key\x18\x0e \x01(\tH\x04\x88\x01\x01\x12\x38\n\x0fscheduled_start\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x05\x88\x01\x01\x42\x12\n\x10_workflow_run_idB\x0c\n\n_parent_idB\x11\n\x0f_attempt_numberB\x12\n\x10_timeout_secondsB\x12\n\x10_idempotency_keyB\x12\n\x10_scheduled_start\"4\n\x0e\x44\x65queueRequest\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t\x12\r\n\x05queue\x18\x02 \x01(\t\"E\n\x0f\x44\x65queueResponse\x12\x32\n\x08task_run\x18\x01 \x01(\x0b\x32 .hyrex.performanceserver.TaskRun\";\n\x10GetQueuesRequest\x12\x16\n\x0emax_num_queues\x18\x01 \x01(\x05\x12\x0f\n\x07pattern\x18\x02 \x01(\t\"4\n\x11GetQueuesResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0e\n\x06queues\x18\x02 \x03(\t\"C\n\x13RetryTaskRunRequest\x12\x13\n\x0btask_run_id\x18\x01 \x01(\t\x12\x17\n\x0f\x62\x61\x63koff_seconds\x18\x02 \x01(\x05\"(\n\x11GetTaskRunRequest\x12\x13\n\x0btask_run_id\x18\x01 \x01(\t\"H\n\x12GetTaskRunResponse\x12\x32\n\x08task_run\x18\x01 \x01(\x0b\x32 .hyrex.performanceserver.TaskRun\".\n\x17GetTaskRunStatusRequest\x12\x13\n\x0btask_run_id\x18\x01 \x01(\t\"O\n\x18GetTaskRunStatusResponse\x12\x33\n\x06status\x18\x01 \x01(\x0e\x32#.hyrex.performanceserver.TaskStatus\"^\n\x17TaskRunHeartbeatRequest\x12\x14\n\x0ctask_run_ids\x18\x01 \x03(\t\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"I\n\x12MarkSuccessRequest\x12\x13\n\x0btask_run_id\x18\x01 \x01(\t\x12\x13\n\x06result\x18\x02 \x01(\tH\x00\x88\x01\x01\x42\t\n\x07_result\"(\n\x11MarkFailedRequest\x12\x13\n\x0btask_run_id\x18\x01 \x01(\t\"/\n\x19GetDurableTaskRunsRequest\x12\x12\n\ndurable_id\x18\x01 \x01(\t\"Q\n\x1aGetDurableTaskRunsResponse\x12\x33\n\ttask_runs\x18\x02 \x03(\x0b\x32 .hyrex.performanceserver.TaskRun\":\n\x11SetLogLinkRequest\x12\x13\n\x0btask_run_id\x18\x01 \x01(\t\x12\x10\n\x08log_link\x18\x02 \x01(\t\"L\n\x16RegisterTaskDefRequest\x12\x32\n\x08task_def\x18\x01 \x01(\x0b\x32 .hyrex.performanceserver.TaskDef\"&\n\x11GetTaskDefRequest\x12\x11\n\ttask_name\x18\x01 \x01(\t\"H\n\x12GetTaskDefResponse\x12\x32\n\x08task_def\x18\x01 \x01(\x0b\x32 .hyrex.performanceserver.TaskDef\"\x17\n\x15GetAllTaskDefsRequest\"M\n\x16GetAllTaskDefsResponse\x12\x33\n\ttask_defs\x18\x01 \x03(\x0b\x32 .hyrex.performanceserver.TaskDef\"\x15\n\x13RegisterCronRequest\"\x81\x01\n\x17RegisterExecutorRequest\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t\x12\x15\n\rexecutor_name\x18\x02 \x01(\t\x12\x15\n\rqueue_pattern\x18\x03 \x01(\t\x12\x0e\n\x06queues\x18\x04 \x03(\t\x12\x13\n\x0bworker_name\x18\x05 \x01(\t\"B\n\x1bUpdateExecutorQueuesRequest\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t\x12\x0e\n\x06queues\x18\x02 \x03(\t\"b\n\x1aUpdateExecutorStatsRequest\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t\x12/\n\x0e\x65xecutor_stats\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"_\n\x18\x45xecutorHeartbeatRequest\x12\x14\n\x0c\x65xecutor_ids\x18\x01 \x03(\t\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"0\n\x19\x44isconnectExecutorRequest\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t\"2\n\x1bMarkRunningTasksLostRequest\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t\"?\n\x12RegisterAppRequest\x12)\n\x08\x61pp_info\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\"5\n\x10WriteLogsRequest\x12\x13\n\x0btask_run_id\x18\x01 \x01(\t\x12\x0c\n\x04logs\x18\x02 \x01(\t\"\x80\x02\n\x17RegisterWorkflowRequest\x12\x15\n\rworkflow_name\x18\x01 \x01(\t\x12\x13\n\x0bsource_code\x18\x02 \x01(\t\x12\x19\n\x11workflow_dag_json\x18\x03 \x01(\t\x12\x39\n\x13workflow_arg_schema\x18\x04 \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x88\x01\x01\x12/\n\x0e\x64\x65\x66\x61ult_config\x18\x05 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x11\n\x04\x63ron\x18\x06 \x01(\tH\x01\x88\x01\x01\x42\x16\n\x14_workflow_arg_schemaB\x07\n\x05_cron\"\xe2\x01\n\x16SendWorkflowRunRequest\x12\x17\n\x0fworkflow_run_id\x18\x01 \x01(\t\x12\x15\n\rworkflow_name\x18\x02 \x01(\t\x12%\n\x04\x61rgs\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\r\n\x05queue\x18\x04 \x01(\t\x12\x1c\n\x0ftimeout_seconds\x18\x05 \x01(\x05H\x00\x88\x01\x01\x12\x1c\n\x0fidempotency_key\x18\x06 \x01(\tH\x01\x88\x01\x01\x42\x12\n\x10_timeout_secondsB\x12\n\x10_idempotency_key\"4\n\x19GetWorkflowRunArgsRequest\x12\x17\n\x0fworkflow_run_id\x18\x01 \x01(\t\"C\n\x1aGetWorkflowRunArgsResponse\x12%\n\x04\x61rgs\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\"4\n\x19\x41\x64vanceWorkflowRunRequest\x12\x17\n\x0fworkflow_run_id\x18\x01 \x01(\t\"8\n\x1dGetWorkflowDurableRunsRequest\x12\x17\n\x0fworkflow_run_id\x18\x01 \x01(\t\"5\n\x1eGetWorkflowDurableRunsResponse\x12\x13\n\x0b\x64urable_ids\x18\x01 \x03(\t\"2\n\x1cTryToCancelDurableRunRequest\x12\x12\n\ndurable_id\x18\x01 \x01(\t\"6\n\x1eGetTaskRunsUpForCancelResponse\x12\x14\n\x0ctask_run_ids\x18\x01 \x03(\t\"*\n\x13MarkCanceledRequest\x12\x13\n\x0btask_run_id\x18\x01 \x01(\t\"U\n\x11KVStoreSetRequest\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x12\x16\n\toverwrite\x18\x03 \x01(\x08H\x00\x88\x01\x01\x42\x0c\n\n_overwrite\" \n\x11KVStoreGetRequest\x12\x0b\n\x03key\x18\x01 \x01(\t\"#\n\x12KVStoreGetResponse\x12\r\n\x05value\x18\x01 \x01(\t\"#\n\x14KVStoreDeleteRequest\x12\x0b\n\x03key\x18\x01 \x01(\tB\x08Z\x06proto/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'requests_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z\006proto/'
  _globals['_ENQUEUEREQUEST']._serialized_start=119
  _globals['_ENQUEUEREQUEST']._serialized_end=651
  _globals['_DEQUEUEREQUEST']._serialized_start=653
  _globals['_DEQUEUEREQUEST']._serialized_end=705
  _globals['_DEQUEUERESPONSE']._serialized_start=707
  _globals['_DEQUEUERESPONSE']._serialized_end=776
  _globals['_GETQUEUESREQUEST']._serialized_start=778
  _globals['_GETQUEUESREQUEST']._serialized_end=837
  _globals['_GETQUEUESRESPONSE']._serialized_start=839
  _globals['_GETQUEUESRESPONSE']._serialized_end=891
  _globals['_RETRYTASKRUNREQUEST']._serialized_start=893
  _globals['_RETRYTASKRUNREQUEST']._serialized_end=960
  _globals['_GETTASKRUNREQUEST']._serialized_start=962
  _globals['_GETTASKRUNREQUEST']._serialized_end=1002
  _globals['_GETTASKRUNRESPONSE']._serialized_start=1004
  _globals['_GETTASKRUNRESPONSE']._serialized_end=1076
  _globals['_GETTASKRUNSTATUSREQUEST']._serialized_start=1078
  _globals['_GETTASKRUNSTATUSREQUEST']._serialized_end=1124
  _globals['_GETTASKRUNSTATUSRESPONSE']._serialized_start=1126
  _globals['_GETTASKRUNSTATUSRESPONSE']._serialized_end=1205
  _globals['_TASKRUNHEARTBEATREQUEST']._serialized_start=1207
  _globals['_TASKRUNHEARTBEATREQUEST']._serialized_end=1301
  _globals['_MARKSUCCESSREQUEST']._serialized_start=1303
  _globals['_MARKSUCCESSREQUEST']._serialized_end=1376
  _globals['_MARKFAILEDREQUEST']._serialized_start=1378
  _globals['_MARKFAILEDREQUEST']._serialized_end=1418
  _globals['_GETDURABLETASKRUNSREQUEST']._serialized_start=1420
  _globals['_GETDURABLETASKRUNSREQUEST']._serialized_end=1467
  _globals['_GETDURABLETASKRUNSRESPONSE']._serialized_start=1469
  _globals['_GETDURABLETASKRUNSRESPONSE']._serialized_end=1550
  _globals['_SETLOGLINKREQUEST']._serialized_start=1552
  _globals['_SETLOGLINKREQUEST']._serialized_end=1610
  _globals['_REGISTERTASKDEFREQUEST']._serialized_start=1612
  _globals['_REGISTERTASKDEFREQUEST']._serialized_end=1688
  _globals['_GETTASKDEFREQUEST']._serialized_start=1690
  _globals['_GETTASKDEFREQUEST']._serialized_end=1728
  _globals['_GETTASKDEFRESPONSE']._serialized_start=1730
  _globals['_GETTASKDEFRESPONSE']._serialized_end=1802
  _globals['_GETALLTASKDEFSREQUEST']._serialized_start=1804
  _globals['_GETALLTASKDEFSREQUEST']._serialized_end=1827
  _globals['_GETALLTASKDEFSRESPONSE']._serialized_start=1829
  _globals['_GETALLTASKDEFSRESPONSE']._serialized_end=1906
  _globals['_REGISTERCRONREQUEST']._serialized_start=1908
  _globals['_REGISTERCRONREQUEST']._serialized_end=1929
  _globals['_REGISTEREXECUTORREQUEST']._serialized_start=1932
  _globals['_REGISTEREXECUTORREQUEST']._serialized_end=2061
  _globals['_UPDATEEXECUTORQUEUESREQUEST']._serialized_start=2063
  _globals['_UPDATEEXECUTORQUEUESREQUEST']._serialized_end=2129
  _globals['_UPDATEEXECUTORSTATSREQUEST']._serialized_start=2131
  _globals['_UPDATEEXECUTORSTATSREQUEST']._serialized_end=2229
  _globals['_EXECUTORHEARTBEATREQUEST']._serialized_start=2231
  _globals['_EXECUTORHEARTBEATREQUEST']._serialized_end=2326
  _globals['_DISCONNECTEXECUTORREQUEST']._serialized_start=2328
  _globals['_DISCONNECTEXECUTORREQUEST']._serialized_end=2376
  _globals['_MARKRUNNINGTASKSLOSTREQUEST']._serialized_start=2378
  _globals['_MARKRUNNINGTASKSLOSTREQUEST']._serialized_end=2428
  _globals['_REGISTERAPPREQUEST']._serialized_start=2430
  _globals['_REGISTERAPPREQUEST']._serialized_end=2493
  _globals['_WRITELOGSREQUEST']._serialized_start=2495
  _globals['_WRITELOGSREQUEST']._serialized_end=2548
  _globals['_REGISTERWORKFLOWREQUEST']._serialized_start=2551
  _globals['_REGISTERWORKFLOWREQUEST']._serialized_end=2807
  _globals['_SENDWORKFLOWRUNREQUEST']._serialized_start=2810
  _globals['_SENDWORKFLOWRUNREQUEST']._serialized_end=3036
  _globals['_GETWORKFLOWRUNARGSREQUEST']._serialized_start=3038
  _globals['_GETWORKFLOWRUNARGSREQUEST']._serialized_end=3090
  _globals['_GETWORKFLOWRUNARGSRESPONSE']._serialized_start=3092
  _globals['_GETWORKFLOWRUNARGSRESPONSE']._serialized_end=3159
  _globals['_ADVANCEWORKFLOWRUNREQUEST']._serialized_start=3161
  _globals['_ADVANCEWORKFLOWRUNREQUEST']._serialized_end=3213
  _globals['_GETWORKFLOWDURABLERUNSREQUEST']._serialized_start=3215
  _globals['_GETWORKFLOWDURABLERUNSREQUEST']._serialized_end=3271
  _globals['_GETWORKFLOWDURABLERUNSRESPONSE']._serialized_start=3273
  _globals['_GETWORKFLOWDURABLERUNSRESPONSE']._serialized_end=3326
  _globals['_TRYTOCANCELDURABLERUNREQUEST']._serialized_start=3328
  _globals['_TRYTOCANCELDURABLERUNREQUEST']._serialized_end=3378
  _globals['_GETTASKRUNSUPFORCANCELRESPONSE']._serialized_start=3380
  _globals['_GETTASKRUNSUPFORCANCELRESPONSE']._serialized_end=3434
  _globals['_MARKCANCELEDREQUEST']._serialized_start=3436
  _globals['_MARKCANCELEDREQUEST']._serialized_end=3478
  _globals['_KVSTORESETREQUEST']._serialized_start=3480
  _globals['_KVSTORESETREQUEST']._serialized_end=3565
  _globals['_KVSTOREGETREQUEST']._serialized_start=3567
  _globals['_KVSTOREGETREQUEST']._serialized_end=3599
  _globals['_KVSTOREGETRESPONSE']._serialized_start=3601
  _globals['_KVSTOREGETRESPONSE']._serialized_end=3636
  _globals['_KVSTOREDELETEREQUEST']._serialized_start=3638
  _globals['_KVSTOREDELETEREQUEST']._serialized_end=3673
# @@protoc_insertion_point(module_scope)



class EnvVars:
    DATABASE_URL = "HYREX_DATABASE_URL"
    API_KEY = "HYREX_API_KEY"
    WORKER_PROCESS = "HYREX_WORKER_PROCESS"
    LOGS_BUCKET = "HYREX_S3_LOG_BUCKET"
    PERF_SERVER_HOST = "HYREX_PERFORMANCE_SERVER_HOST"
    PERF_SERVER_PORT = "HYREX_PERFORMANCE_SERVER_PORT"
    LOCAL_TESTING = "HYREX_LOCAL_TESTING"



import asyncio
import importlib
import json
import logging
import os
import random
import signal
import socket
import sys
import time
import traceback
from datetime import datetime, timezone
from inspect import signature
from multiprocessing import Event, Process, Queue
from pathlib import Path
from uuid import UUID

from psycopg.types.json import Json
from pydantic import BaseModel

from hyrex import constants
from hyrex.dispatcher import DequeuedTask, get_dispatcher
from hyrex.dispatcher.performance_dispatcher import PerformanceDispatcher
from hyrex.env_vars import EnvVars
from hyrex.hyrex_app import HyrexApp, HyrexAppInfo
from hyrex.hyrex_cache import HyrexCacheManager
from hyrex.hyrex_context import HyrexContext, clear_hyrex_context, set_hyrex_context
from hyrex.hyrex_queue import HyrexQueue
from hyrex.hyrex_registry import HyrexRegistry
from hyrex.schemas import QueuePattern
from hyrex.worker.executor.time_series_averager import TimeSeriesAverager
from hyrex.worker.logging import LogLevel, init_logging
from hyrex.worker.messages.root_messages import (
    SetExecutorTaskMessage,
    TaskRegistrationComplete,
)
from hyrex.worker.s3_logs import write_task_logs_to_s3, write_task_logs_with_dispatcher
from hyrex.worker.utils import glob_pattern_to_postgres_pattern, is_glob_pattern, is_process_alive


def generate_executor_name():
    hostname = socket.gethostname()
    pid = os.getpid()
    timestamp = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    return f"hyrex-executor-{hostname}-{pid}-{timestamp}"


class HyrexTaskTimeout(Exception):
    pass


class WorkerExecutor(Process):

    def __init__(
        self,
        root_message_queue: Queue,
        log_level: LogLevel,
        app_module_path: str,
        executor_id: UUID,
        queue: str,
        executor_name: str,
        worker_name: str,
        register_app: bool = False,
    ):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self.log_level = log_level

        self.root_message_queue = root_message_queue
        self._stop_event = Event()

        self.app_module_path = app_module_path
        self.queue = queue
        self.queue_pattern = None
        self.queues: list[HyrexQueue] = []
        self.executor_id = executor_id
        self.name = executor_name

        self.logs_s3_bucket = os.environ.get(EnvVars.LOGS_BUCKET)

        # Perf metrics
        self.num_distinct_queues_averager = TimeSeriesAverager()
        self.refresh_queue_duration_averager = TimeSeriesAverager()
        self.dequeue_duration_averager = TimeSeriesAverager()

        self.registry: HyrexRegistry = None
        self.register_app = register_app

        # To check if root process is running
        self.parent_pid = os.getpid()
        self.worker_name = worker_name

    def get_concurrency_for_queue(self, queue_name: str):
        return self.registry.get_concurrency_limit(queue_name=queue_name)

    def update_queue_list(self):
        self.queues = []
        self.logger.debug("Updating internal queue list from pattern...")

        start = time.perf_counter()
        queue_names = self.dispatcher.get_queues_for_pattern(self.queue_pattern)
        end = time.perf_counter()
        # Milliseconds
        self.refresh_queue_duration_averager.submit((end - start) * 1000)
        self.num_distinct_queues_averager.submit(len(queue_names))

        self.logger.debug(f"Queues found: {queue_names}")
        if queue_names:
            random.shuffle(queue_names)
        else:
            self._stop_event.wait(1.0)
            return

        for queue_name in queue_names:
            self.queues.append(
                HyrexQueue(
                    name=queue_name,
                    concurrency_limit=self.get_concurrency_for_queue(
                        queue_name=queue_name
                    ),
                )
            )

        self.dispatcher.update_executor_queues(self.executor_id, queue_names)

    def load_app_module(self):
        sys.path.append(str(Path.cwd()))
        module_path, instance_name = self.app_module_path.split(":")
        # Import the worker module
        app_module = importlib.import_module(module_path)
        app_instance: HyrexApp = getattr(app_module, instance_name)

        self.registry = app_instance.registry
        self.app_info = app_instance.app_info

    async def process_item(self, task: DequeuedTask):
        task_wrapper = self.registry.get_task(task.task_name)

        # Write logs via performance server
        if isinstance(self.dispatcher, PerformanceDispatcher):
            async with write_task_logs_with_dispatcher(
                task_id=task.id, dispatcher=self.dispatcher
            ):
                result = await task_wrapper.async_call(**task.args)
        # Write logs directly to S3
        elif self.logs_s3_bucket:
            try:
                async with write_task_logs_to_s3(
                    task.id, self.logs_s3_bucket
                ) as s3_log_link:
                    result = await task_wrapper.async_call(**task.args)
            finally:
                self.dispatcher.set_log_link(task.id, s3_log_link)
        # No logs written to S3
        else:
            result = await task_wrapper.async_call(**task.args)

        return result

    def fetch_task(self, queue: str, concurrency_limit: int = 0) -> DequeuedTask | None:
        start = time.perf_counter()
        dequeued_task = self.dispatcher.dequeue(
            executor_id=self.executor_id,
            # TODO: Cache this list.
            task_names=self.registry.get_task_names(),
            queue=queue,
            concurrency_limit=concurrency_limit,
        )
        end = time.perf_counter()
        # Milliseconds
        self.dequeue_duration_averager.submit((end - start) * 1000)
        return dequeued_task

    def mark_task_success(self, task_id: UUID, result: str):
        self.dispatcher.mark_success(task_id=task_id, result=result)

    def mark_task_failed(self, task_id: UUID):
        self.dispatcher.mark_failed(task_id=task_id)

    def retry_task(self, task_id: UUID, backoff_seconds: int):
        self.dispatcher.retry_task(task_id=task_id, backoff_seconds=backoff_seconds)

    # Notifies root process of current task being processed.
    def update_current_task(self, task_id: UUID):
        self.root_message_queue.put(
            SetExecutorTaskMessage(executor_id=self.executor_id, task_id=task_id),
        )

    def process(self, queue: HyrexQueue) -> bool:
        """Returns True if a task is found and attempted, False otherwise"""
        task: DequeuedTask | None = self.fetch_task(
            queue=queue.name, concurrency_limit=queue.concurrency_limit
        )
        if not task:
            return False

        try:
            set_hyrex_context(
                HyrexContext(
                    task_id=task.id,
                    durable_id=task.durable_id,
                    root_id=task.root_id,
                    parent_id=task.parent_id,
                    task_name=task.task_name,
                    queue=task.queue,
                    priority=task.priority,
                    timeout_seconds=task.timeout_seconds,
                    scheduled_start=task.scheduled_start,
                    queued=task.queued,
                    started=task.started,
                    executor_id=self.executor_id,
                    attempt_number=task.attempt_number,
                    max_retries=task.max_retries,
                    workflow_run_id=task.workflow_run_id,
                )
            )

            # Notify root process of new task
            self.update_current_task(task.id)
            # Set up timeout
            if task.timeout_seconds:
                signal.alarm(task.timeout_seconds)
            # Run task
            result = asyncio.run(self.process_item(task))

            if result is not None:
                if isinstance(result, BaseModel):
                    result = result.model_dump_json()
                else:
                    try:
                        result = json.dumps(result)
                    except (TypeError, ValueError):
                        raise TypeError("Return value must be JSON-serializable")

            self.mark_task_success(task.id, result)

            self.logger.info(
                f"Executor {self.name}: Completed processing item {task.id}"
            )

            # If this task is part of a workflow, advance it
            if task.workflow_run_id:
                self.logger.info(f"Advancing workflow {task.workflow_run_id}...")
                self.dispatcher.advance_workflow_run(
                    workflow_run_id=task.workflow_run_id
                )

        except Exception as e:
            self.logger.error(f"Executor {self.name}: Exception hit during processing.")
            self.logger.error(e)
            self.logger.error("Traceback:\n%s", traceback.format_exc())

            if "task" in locals():
                self.logger.error(f"Marking task {task.id} as failed.")
                self.mark_task_failed(task.id)
                if task.attempt_number < task.max_retries:
                    self.logger.info("Submitting task for retry...")
                    try:
                        backoff_seconds = self.registry.get_retry_backoff(
                            task_name=task.task_name, attempt_number=task.attempt_number
                        )
                        self.retry_task(
                            task_id=task.id, backoff_seconds=backoff_seconds
                        )
                    except Exception as retry_error:
                        self.logger.error(f"Error during retry process: {retry_error}")
                        self.logger.error(
                            f"Retry error traceback:\n%s", traceback.format_exc()
                        )

                on_error = self.registry.get_on_error_handler(task.task_name)
                if on_error:
                    self.logger.info(
                        f"Running on_error handler for task {task.task_name}"
                    )
                    try:
                        sig = signature(on_error)
                        if len(sig.parameters) == 0:
                            on_error()
                        else:
                            on_error(e)

                    except Exception as on_error_exception:
                        self.logger.error(
                            "Exception hit when running on_error handler."
                        )
                        self.logger.error(on_error_exception)
                        self.logger.error("Traceback:\n%s", traceback.format_exc())

            self._stop_event.wait(0.5)  # Add delay after error
            return True
        finally:
            self.update_current_task(None)
            signal.alarm(0)  # Clear alarm
            clear_hyrex_context()

            # 1/25 chance to publish stats
            if random.random() < 0.04:
                stats = {
                    "dequeueLatencyMs": [
                        avg.model_dump()
                        for avg in self.dequeue_duration_averager.get_time_series()
                    ],
                    "refreshQueueLatencyMs": [
                        avg.model_dump()
                        for avg in self.refresh_queue_duration_averager.get_time_series()
                    ],
                    "numDistinctQueues": [
                        avg.model_dump()
                        for avg in self.num_distinct_queues_averager.get_time_series()
                    ],
                }
                self.dispatcher.update_executor_stats(self.executor_id, stats)

            return True

    def check_root_process(self):
        # Confirm parent is still alive
        if not is_process_alive(self.parent_pid):
            self.logger.warning("Root process died unexpectedly. Shutting down.")
            self._stop_event.set()

    def run_static_queue_loop(self):
        queue = HyrexQueue(
            name=self.queue,
            concurrency_limit=self.registry.get_concurrency_limit(
                queue_name=self.queue
            ),
        )
        while not self._stop_event.is_set():
            # Queue pattern is a static string - fetch from it directly.
            if not self.process(queue):
                # No task found, sleep for a bit
                self._stop_event.wait(0.5)
            self.check_root_process()

    def run_round_robin_loop(self):
        self.update_queue_list()
        last_queue_refresh = time.monotonic()
        no_task_count = 0

        while not self._stop_event.is_set():
            seconds_since_queue_refresh = time.monotonic() - last_queue_refresh
            if (
                seconds_since_queue_refresh
                > constants.WORKER_EXECUTOR_QUEUE_REFRESH_SECONDS
                or len(self.queues) == 0
                or no_task_count >= 5
            ):
                self.update_queue_list()
                last_queue_refresh = time.monotonic()
                no_task_count = 0

            for queue in self.queues:
                self.check_root_process()
                if self._stop_event.is_set():
                    break

                if not self.process(queue=queue):
                    no_task_count += 1
                else:
                    no_task_count = 0

                # We're not hitting populated queues - pause and refresh queue list.
                if no_task_count >= 5:
                    break

    def register_hyrex_app(self):
        self.dispatcher.register_app(self.app_info.model_dump())

    def run(self):
        init_logging(self.log_level)

        # Retrieve name and task registry from the provided app module path.
        self.load_app_module()

        # Convert queue pattern to Postgres SIMILAR TO pattern if needed.
        if is_glob_pattern(self.queue):
            self.queue_pattern = QueuePattern(
                glob_pattern=self.queue,
                postgres_pattern=glob_pattern_to_postgres_pattern(self.queue),
            )
            self.logger.debug(
                f"Converted queue glob to Postgres SIMILAR TO pattern: {self.queue_pattern.glob_pattern} -> {self.queue_pattern.postgres_pattern}"
            )

        try:
            self.dispatcher = get_dispatcher()
            self.dispatcher.register_executor(
                executor_id=self.executor_id,
                executor_name=self.name,
                queue_pattern=self.queue,
                queues=self.queues,
                worker_name=self.worker_name,
            )

            # Ignore termination signals, let main process manage shutdown.
            signal.signal(signal.SIGTERM, signal.SIG_IGN)
            signal.signal(signal.SIGINT, signal.SIG_IGN)

            if self.register_app:
                self.logger.info(
                    f"{self.name}: Registering app, tasks, and workflows to the DB."
                )
                self.registry.register_all_with_db()
                self.register_hyrex_app()
                # Notify root process that cron scheduler can start
                self.root_message_queue.put(TaskRegistrationComplete())

            # Set up to throw HyrexTaskTimeout and then end process on task timeout.
            def timeout_handler(signum, frame):
                self._stop_event.set()
                raise HyrexTaskTimeout()

            signal.signal(signal.SIGALRM, timeout_handler)

            self.logger.info(
                f"Executor process {self.name} started - checking for tasks."
            )

            # Run the main loop
            if self.queue_pattern:
                self.run_round_robin_loop()
            else:
                self.run_static_queue_loop()
        finally:
            self.stop()

    def stop(self):
        self.logger.info(f"Stopping {self.name}...")
        if self.dispatcher:
            self.dispatcher.disconnect_executor(self.executor_id)
            self.dispatcher.stop()
        # Clean up any cached resources
        HyrexCacheManager.cleanup()
        self.logger.info(f"{self.name} stopped successfully!")



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: fill_historical_task_status_counts_table.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


FILL_HISTORICAL_TASK_STATUS_COUNTS_TABLE = """-- name: fill_historical_task_status_counts_table \\:exec
WITH RECURSIVE timepoints AS (
        -- 1) Start from the larger of\\:
        --    - The last known timepoint from the stats table (if any)
        --    - 10 minutes ago (rounded to a 15s boundary)
        SELECT GREATEST(
                       COALESCE(
                               (SELECT MAX(timepoint) FROM hyrex_stats_task_status_counts),
                               date_bin(
                                       INTERVAL '15 seconds',
                                       now() - INTERVAL '10 minutes',
                                       TIMESTAMP '2000-01-01 00\\:00\\:00+00'
                               )
                       ),
                       date_bin(
                               INTERVAL '15 seconds',
                               now() - INTERVAL '10 minutes',
                               TIMESTAMP '2000-01-01 00\\:00\\:00+00'
                       )
               ) + INTERVAL '15 seconds' AS timepoint

        UNION ALL

        -- 2) Keep adding 15 seconds, up to 'now' (also rounded to a 15s boundary)
        SELECT timepoint + INTERVAL '15 seconds'
        FROM timepoints
        WHERE timepoint < date_bin(
                INTERVAL '15 seconds',
                now(),
                TIMESTAMP '2000-01-01 00\\:00\\:00+00'
                          )
    ),
               queue_counts AS (
                   SELECT
                       t.timepoint,
                       COUNT(
                               CASE
                                   WHEN htr.queued <= t.timepoint
                                       AND (htr.started IS NULL OR htr.started > t.timepoint)
                                       AND htr.status = 'QUEUED'\\:\\:task_run_status
                                       THEN 1
                                   END
                       ) AS queued,
                       COUNT(
                               CASE
                                   WHEN htr.started <= t.timepoint
                                       AND (htr.finished IS NULL OR htr.finished > t.timepoint)
                                       AND htr.status = 'RUNNING'\\:\\:task_run_status
                                       THEN 1
                                   END
                       ) AS running,
                       COUNT(
                               CASE
                                   WHEN htr.status IN ('AWAIT_DEPS'\\:\\:task_run_status, 'AWAIT_START_TIME'\\:\\:task_run_status)
                                       AND htr.queued <= t.timepoint
                                       AND (htr.finished IS NULL OR htr.finished > t.timepoint)
                                       THEN 1
                                   END
                       ) AS waiting,
                       COUNT(
                               CASE
                                   WHEN htr.status = 'FAILED'\\:\\:task_run_status
                                       AND htr.finished <= t.timepoint
                                       THEN 1
                                   END
                       ) AS failed,
                       COUNT(
                               CASE
                                   WHEN htr.status = 'SUCCESS'\\:\\:task_run_status
                                       AND htr.finished <= t.timepoint
                                       THEN 1
                                   END
                       ) AS success,
                       COUNT(
                               CASE
                                   WHEN htr.status = 'LOST'\\:\\:task_run_status
                                       AND (htr.finished IS NULL OR htr.finished <= t.timepoint)
                                       THEN 1
                                   END
                       ) AS lost
                   FROM timepoints t
                            LEFT JOIN hyrex_task_run htr
                                      ON (
                                          -- Include tasks that existed during this timepoint
                                          htr.queued <= t.timepoint
                                              AND (
                                              -- Either they're still in the system
                                              htr.finished IS NULL
                                                  OR
                                                  -- Or they finished after this timepoint
                                              htr.finished > t.timepoint
                                                  OR
                                                  -- Or they failed/succeeded/lost at this exact timepoint
                                              (
                                                  htr.status IN ('FAILED'\\:\\:task_run_status,'SUCCESS'\\:\\:task_run_status,'LOST'\\:\\:task_run_status)
                                                      AND htr.finished <= t.timepoint
                                                  )
                                              )
                                          )
                   GROUP BY t.timepoint
               ),
               final_counts AS (
                   SELECT
                       timepoint,
                       queued,
                       running,
                       waiting,
                       failed,
                       success,
                       lost,
                       (queued + running + waiting + failed + lost) AS total,
                       (queued - LAG(queued, 1) OVER (ORDER BY timepoint))   AS queued_delta,
                       (success - LAG(success, 1) OVER (ORDER BY timepoint)) AS success_delta,
                       (failed - LAG(failed, 1) OVER (ORDER BY timepoint))   AS failed_delta,
                       (lost - LAG(lost, 1) OVER (ORDER BY timepoint))       AS lost_delta
                   FROM queue_counts
               )
INSERT INTO hyrex_stats_task_status_counts
SELECT
    timepoint,
    queued,
    running,
    waiting,
    failed,
    success,
    lost,
    total,
    queued_delta,
    success_delta,
    failed_delta,
    lost_delta
FROM final_counts
WHERE queued_delta IS NOT NULL
  AND success_delta IS NOT NULL
  AND failed_delta IS NOT NULL
  AND lost_delta IS NOT NULL
ON CONFLICT (timepoint) DO NOTHING
"""


@dataclasses.dataclass()
class FillHistoricalTaskStatusCountsTableParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def fill_historical_task_status_counts_table(self, arg: FillHistoricalTaskStatusCountsTableParams) -> None:
        self._conn.execute(sqlalchemy.text(FILL_HISTORICAL_TASK_STATUS_COUNTS_TABLE))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def fill_historical_task_status_counts_table(self, arg: FillHistoricalTaskStatusCountsTableParams) -> None:
        await self._conn.execute(sqlalchemy.text(FILL_HISTORICAL_TASK_STATUS_COUNTS_TABLE))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: set_workflow_run_status_based_on_task_runs.sql
import dataclasses
from typing import Any, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


SET_WORKFLOW_RUN_STATUS_BASED_ON_TASK_RUNS = """-- name: set_workflow_run_status_based_on_task_runs \\:one
SELECT set_workflow_run_status_based_on_task_runs FROM set_workflow_run_status_based_on_task_runs(:p1)
"""


@dataclasses.dataclass()
class SetWorkflowRunStatusBasedOnTaskRunsParams:
    workflow_run_id: uuid.UUID


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def set_workflow_run_status_based_on_task_runs(self, arg: SetWorkflowRunStatusBasedOnTaskRunsParams) -> Optional[Optional[Any]]:
        row = self._conn.execute(sqlalchemy.text(SET_WORKFLOW_RUN_STATUS_BASED_ON_TASK_RUNS), {"p1": arg.workflow_run_id}).first()
        if row is None:
            return None
        return row[0]


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def set_workflow_run_status_based_on_task_runs(self, arg: SetWorkflowRunStatusBasedOnTaskRunsParams) -> Optional[Optional[Any]]:
        row = (await self._conn.execute(sqlalchemy.text(SET_WORKFLOW_RUN_STATUS_BASED_ON_TASK_RUNS), {"p1": arg.workflow_run_id})).first()
        if row is None:
            return None
        return row[0]



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: insert_workflow_run.sql
import dataclasses
from typing import Any, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


INSERT_WORKFLOW_RUN = """-- name: insert_workflow_run \\:one
INSERT INTO hyrex_workflow_run (
    id,
    parent_id,
    workflow_name,
    args,
    queue,
    timeout_seconds,
    status,
    queued,
    started,
    last_heartbeat,
    idempotency_key
)
VALUES (:p1, NULL, :p2, :p3, :p4, :p5, 'RUNNING'\\:\\:workflow_run_status, now(), now(), now(), :p6)
RETURNING id
"""


@dataclasses.dataclass()
class InsertWorkflowRunParams:
    id: uuid.UUID
    workflow_name: str
    args: Any
    queue: str
    timeout_seconds: Optional[int]
    idempotency_key: Optional[str]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def insert_workflow_run(self, arg: InsertWorkflowRunParams) -> Optional[uuid.UUID]:
        row = self._conn.execute(sqlalchemy.text(INSERT_WORKFLOW_RUN), {
            "p1": arg.id,
            "p2": arg.workflow_name,
            "p3": arg.args,
            "p4": arg.queue,
            "p5": arg.timeout_seconds,
            "p6": arg.idempotency_key,
        }).first()
        if row is None:
            return None
        return row[0]


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def insert_workflow_run(self, arg: InsertWorkflowRunParams) -> Optional[uuid.UUID]:
        row = (await self._conn.execute(sqlalchemy.text(INSERT_WORKFLOW_RUN), {
            "p1": arg.id,
            "p2": arg.workflow_name,
            "p3": arg.args,
            "p4": arg.queue,
            "p5": arg.timeout_seconds,
            "p6": arg.idempotency_key,
        })).first()
        if row is None:
            return None
        return row[0]



import json
import threading
import time
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from queue import Empty, Queue
from typing import List, Type
from uuid import UUID

import sqlalchemy
from pydantic import BaseModel
from sqlalchemy import create_engine
from uuid6 import uuid7

from hyrex import constants
from hyrex.configs import TaskConfig
from hyrex.dispatcher.dispatcher import Dispatcher
from hyrex.hyrex_queue import HyrexQueue
from hyrex.schemas import (
    CronJob,
    CronJobRun,
    DequeuedTask,
    EnqueueTaskRequest,
    QueuePattern,
    TaskRun,
    TaskStatus,
    WorkflowRunRequest,
    WorkflowStatus,
)

# Import SQLC generated models and parameter classes
from .sqlc import models
from .sqlc import (
    # Import modules containing parameter classes
    register_app_info,
    transition_task_state,
    save_result,
    set_log_link,
    conditionally_retry_task,
    fetch_task,
    fetch_task_with_concurrency_limit,
    create_task_run,
    get_task_run_by_id,
    register_executor,
    disconnect_executor,
    batch_update_heartbeat_on_executors,
    update_executor_stats,
    batch_update_heartbeat_log,
    fetch_active_queue_names,
    register_task_def,
    create_cron_job_for_task,
    turn_off_cron_for_task,
    register_workflow,
    create_workflow_run,
    set_workflow_run_status_based_on_task_runs,
    advance_workflow_run_func,
    get_workflow_run_by_id,
    acquire_scheduler_lock,
    pull_active_cron_expressions,
    update_cron_job_confirmation_ts,
    schedule_cron_job_runs_json,
    trigger_execute_queued_cron_job,
    create_cron_job_for_sql_query,
    release_scheduler_lock,
    get_task_attempts_by_durable_id,
    get_workflow_run_task_runs,
    update_queues_on_executor,
    fetch_result,
    kv_set_value,
    kv_get_value,
    kv_delete_value,
    # Import sync functions
    register_app_info_sync,
    transition_task_state_sync,
    save_result_sync,
    set_log_link_sync,
    conditionally_retry_task_sync,
    fetch_task_sync,
    fetch_task_with_concurrency_limit_sync,
    create_task_run_sync,
    get_task_run_by_id_sync,
    register_executor_sync,
    disconnect_executor_sync,
    batch_update_heartbeat_on_executors_sync,
    update_executor_stats_sync,
    batch_update_heartbeat_log_sync,
    fetch_active_queue_names_sync,
    register_task_def_sync,
    create_cron_job_for_task_sync,
    turn_off_cron_for_task_sync,
    register_workflow_sync,
    create_workflow_run_sync,
    set_workflow_run_status_based_on_task_runs_sync,
    advance_workflow_run_func_sync,
    get_workflow_run_by_id_sync,
    acquire_scheduler_lock_sync,
    pull_active_cron_expressions_sync,
    update_cron_job_confirmation_ts_sync,
    schedule_cron_job_runs_json_sync,
    trigger_execute_queued_cron_job_sync,
    create_cron_job_for_sql_query_sync,
    release_scheduler_lock_sync,
    get_task_attempts_by_durable_id_sync,
    get_workflow_run_task_runs_sync,
    update_queues_on_executor_sync,
    fetch_result_sync,
    set_value_sync,
    get_value_sync,
    delete_value_sync,
)


def pydantic_aware_default(obj):
    if isinstance(obj, BaseModel):
        return obj.model_dump()
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")


class SqlcDispatcher(Dispatcher):
    """
    A dispatcher implementation that uses SQLC-generated queries with SQLAlchemy.
    """

    def __init__(self, conn_string: str, batch_size=1000, flush_interval=0.05):
        super().__init__()
        self.conn_string = conn_string

        # Create SQLAlchemy engine with psycopg3
        # Replace postgresql:// with postgresql+psycopg:// to use psycopg3
        if conn_string.startswith("postgresql://"):
            conn_string = conn_string.replace(
                "postgresql://", "postgresql+psycopg://", 1
            )
        elif conn_string.startswith("postgres://"):
            conn_string = conn_string.replace("postgres://", "postgresql+psycopg://", 1)

        self.engine = create_engine(
            conn_string,
            pool_size=5,
            max_overflow=10,
            pool_pre_ping=True,
            connect_args={
                "keepalives": 1,
                "keepalives_idle": 60,
                "keepalives_interval": 10,
            },
        )

        self.local_queue = Queue()
        self.batch_size = batch_size
        self.flush_interval = flush_interval

        self.thread = threading.Thread(target=self._batch_enqueue, daemon=True)
        self.thread.start()
        self.stopping = False

        self.register_shutdown_handlers()

    @contextmanager
    def transaction(self):
        """Context manager for database transactions."""
        with self.engine.begin() as conn:
            yield conn

    def register_app(self, app_info: dict):
        with self.transaction() as conn:
            register_app_info_sync(
                conn,
                register_app_info.RegisterAppInfoParams(
                    id=1, app_info=json.dumps(app_info)
                ),
            )

    def mark_success(self, task_id: UUID, result: str):
        with self.transaction() as conn:
            # Transition task state to success
            transition_task_state_sync(
                conn,
                transition_task_state.TransitionTaskStateParams(
                    task_id=task_id, next_state="SUCCESS"
                ),
            )
            # Save result if provided
            if result:
                save_result_sync(
                    conn, save_result.SaveResultParams(task_id=task_id, result=result)
                )

    def mark_failed(self, task_id: UUID):
        with self.transaction() as conn:
            transition_task_state_sync(
                conn,
                transition_task_state.TransitionTaskStateParams(
                    task_id=task_id, next_state="FAILED"
                ),
            )

    def set_log_link(self, task_id: UUID, log_link: str):
        with self.transaction() as conn:
            set_log_link_sync(
                conn, set_log_link.SetLogLinkParams(task_id=task_id, log_link=log_link)
            )

    def retry_task(self, task_id: UUID, backoff_seconds: int):
        scheduled_start = None
        if backoff_seconds > 0:
            scheduled_start = datetime.now(timezone.utc) + timedelta(
                seconds=backoff_seconds
            )

        with self.transaction() as conn:
            conditionally_retry_task_sync(
                conn,
                conditionally_retry_task.ConditionallyRetryTaskParams(
                    task_id=task_id,
                    new_task_id=uuid7(),
                    scheduled_start=scheduled_start,
                ),
            )

    def try_to_cancel_task(self, task_id: UUID):
        with self.transaction() as conn:
            transition_task_state_sync(
                conn,
                transition_task_state.TransitionTaskStateParams(
                    task_id=task_id, next_state="UP_FOR_CANCEL"
                ),
            )

    def task_canceled(self, task_id: UUID):
        with self.transaction() as conn:
            transition_task_state_sync(
                conn,
                transition_task_state.TransitionTaskStateParams(
                    task_id=task_id, next_state="CANCELED"
                ),
            )

    def dequeue(
        self,
        executor_id: UUID,
        task_names: list[str],
        queue: str = constants.ANY_QUEUE,
        concurrency_limit: int = 0,
    ) -> DequeuedTask:
        dequeued_task = None
        with self.transaction() as conn:
            if concurrency_limit > 0:
                row = fetch_task_with_concurrency_limit_sync(
                    conn,
                    fetch_task_with_concurrency_limit.FetchTaskWithConcurrencyLimitParams(
                        queue=queue,
                        concurrency_limit=concurrency_limit,
                        executor_id=executor_id,
                        task_names=task_names,
                    ),
                )
            else:
                row = fetch_task_sync(
                    conn,
                    fetch_task.FetchTaskParams(
                        executor_id=executor_id, queue=queue, task_names=task_names
                    ),
                )

            if row:
                dequeued_task = DequeuedTask(
                    id=row.id,
                    durable_id=row.durable_id,
                    root_id=row.root_id,
                    parent_id=row.parent_id,
                    task_name=row.task_name,
                    args=row.args,
                    queue=row.queue,
                    priority=row.priority,
                    timeout_seconds=row.timeout_seconds,
                    scheduled_start=row.scheduled_start,
                    queued=row.queued,
                    started=row.started,
                    workflow_run_id=row.workflow_run_id,
                    attempt_number=row.attempt_number,
                    max_retries=row.max_retries,
                )

        return dequeued_task

    def enqueue(self, tasks: list[EnqueueTaskRequest]):
        if not tasks:
            self.logger.error("Task list is empty - cannot enqueue.")
            return
        if self.stopping:
            self.logger.warning("Task enqueued during shutdown. May not be processed.")
        for task in tasks:
            self.local_queue.put(task)

    def _batch_enqueue(self):
        tasks = []
        last_flush_time = time.monotonic()

        while True:
            try:
                task = self.local_queue.get(timeout=self.flush_interval)
                if task is None:  # Stop signal
                    break

                tasks.append(task)
                # Drain queue without blocking
                while len(tasks) < self.batch_size:
                    try:
                        task = self.local_queue.get_nowait()
                        if task is None:
                            break
                        tasks.append(task)
                    except Empty:
                        break

                # Check if we should flush
                current_time = time.monotonic()
                if (current_time - last_flush_time >= self.flush_interval) or len(
                    tasks
                ) >= self.batch_size:
                    if tasks:
                        self._enqueue_tasks(tasks)
                        tasks = []
                    last_flush_time = current_time

            except Empty:
                # Flush on timeout if we have tasks
                if tasks:
                    self._enqueue_tasks(tasks)
                    tasks = []
                last_flush_time = time.monotonic()

    def _enqueue_tasks(self, tasks: List[EnqueueTaskRequest]):
        """Inserts a batch of tasks into the database using SQLC."""
        with self.transaction() as conn:
            for task in tasks:
                # Convert args to JSON, handling Pydantic models
                try:
                    args_json = json.dumps(task.args, default=pydantic_aware_default)
                except TypeError as e:
                    self.logger.error(
                        f"Task {task.id}: Failed to serialize args to JSON: {e}"
                    )
                    raise

                # Create task using SQLC generated function
                create_task_run_sync(
                    conn,
                    create_task_run.CreateTaskRunParams(
                        id=task.id,
                        durable_id=task.durable_id,
                        root_id=task.root_id,
                        parent_id=task.parent_id,
                        status=task.status.value,  # Convert enum to string
                        task_name=task.task_name,
                        args=args_json,  # Pass JSON string directly
                        queue=task.queue,
                        max_retries=task.max_retries,
                        priority=task.priority,
                        timeout_seconds=task.timeout_seconds,
                        idempotency_key=task.idempotency_key,
                        scheduled_start=datetime.now(
                            timezone.utc
                        ),  # Use current time for scheduled_start
                        workflow_run_id=task.workflow_run_id,
                        workflow_dependencies=task.workflow_dependencies or [],
                    ),
                )

    def stop(self, timeout: float = 5.0) -> bool:
        # Check if already stopping/stopped
        if self.stopping:
            return True

        self.logger.debug("Stopping dispatcher...")
        self.stopping = True

        # Signal the batch thread to stop and wait with timeout
        self.local_queue.put(None)
        self.thread.join(timeout=timeout)

        clean_shutdown = not self.thread.is_alive()

        # Close the engine
        if clean_shutdown:
            self.engine.dispose()
        else:
            self.logger.warning(
                "Batch thread did not stop cleanly, forcing engine to close"
            )
            self.engine.dispose()

        self.logger.debug(
            "Dispatcher stopped %s.",
            "successfully" if clean_shutdown else "with timeout",
        )
        return clean_shutdown

    def get_task_status(self, task_id: UUID) -> TaskStatus:
        with self.transaction() as conn:
            task_run = get_task_run_by_id_sync(
                conn, get_task_run_by_id.GetTaskRunByIdParams(id=task_id)
            )
            if task_run is None:
                raise ValueError(f"Task id {task_id} not found in DB.")
            return TaskStatus(task_run.status)

    def register_executor(
        self,
        executor_id: UUID,
        executor_name: str,
        queue_pattern: str,
        queues: list[HyrexQueue],
        worker_name: str,
    ):
        with self.transaction() as conn:
            register_executor_sync(
                conn,
                register_executor.RegisterExecutorParams(
                    id=executor_id,
                    name=executor_name,
                    queue_pattern=queue_pattern,
                    queues=[q.name for q in queues],
                    worker_name=worker_name,
                ),
            )

    def disconnect_executor(self, executor_id: UUID):
        with self.transaction() as conn:
            disconnect_executor_sync(
                conn,
                disconnect_executor.DisconnectExecutorParams(
                    stats=None, id=executor_id  # No stats on disconnect
                ),
            )

    def executor_heartbeat(self, executor_ids: list[UUID], timestamp: datetime):
        with self.transaction() as conn:
            batch_update_heartbeat_on_executors_sync(
                conn,
                batch_update_heartbeat_on_executors.BatchUpdateHeartbeatOnExecutorsParams(
                    executor_ids=executor_ids
                ),
            )

    def update_executor_stats(self, executor_id: UUID, stats: dict):
        with self.transaction() as conn:
            update_executor_stats_sync(
                conn,
                update_executor_stats.UpdateExecutorStatsParams(
                    id=executor_id, stats=json.dumps(stats)
                ),
            )

    def task_heartbeat(self, task_ids: list[UUID], timestamp: datetime):
        # TODO: Implement task heartbeat functionality
        # The current batch_update_heartbeat_log is for executor heartbeats, not tasks
        pass

    def get_tasks_up_for_cancel(self) -> list[UUID]:
        # This needs to be implemented based on business logic
        # For now, return empty list
        return []

    def get_queues_for_pattern(self, pattern: QueuePattern) -> list[str]:
        with self.transaction() as conn:
            return list(
                fetch_active_queue_names_sync(
                    conn,
                    fetch_active_queue_names.FetchActiveQueueNamesParams(
                        queue_pattern=pattern.postgres_pattern
                    ),
                )
            )

    def register_task_def(
        self,
        task_name: str,
        arg_schema: Type[BaseModel] | None,
        task_config: TaskConfig,
        cron: str = None,
        source_code: str = None,
    ):
        with self.transaction() as conn:
            # Convert arg_schema to dict if it's a Pydantic model class
            arg_schema_dict = None
            if arg_schema and hasattr(arg_schema, "model_json_schema"):
                arg_schema_dict = json.dumps(arg_schema.model_json_schema())

            register_task_def_sync(
                conn,
                register_task_def.RegisterTaskDefParams(
                    task_name=task_name,
                    cron_expr=cron,
                    source_code=source_code,
                    arg_schema=arg_schema_dict,
                    queue=(
                        task_config.get_queue_name() if task_config.queue else "default"
                    ),
                    priority=(
                        task_config.priority if task_config.priority is not None else 5
                    ),
                    max_retries=(
                        task_config.max_retries
                        if task_config.max_retries is not None
                        else 0
                    ),
                    timeout_seconds=task_config.timeout_seconds,
                ),
            )

            # Handle cron job creation/deletion
            cron_job_name = f"ScheduledTask-{task_name}"
            if cron:
                # Create cron job for task
                current_id = uuid7()
                task_request = EnqueueTaskRequest(
                    id=current_id,
                    durable_id=current_id,
                    workflow_run_id=None,
                    workflow_dependencies=None,
                    root_id=current_id,
                    parent_id=None,
                    queue=(
                        task_config.get_queue_name() if task_config.queue else "default"
                    ),
                    status=TaskStatus.queued,
                    task_name=task_name,
                    args={},
                    max_retries=(
                        task_config.max_retries
                        if task_config.max_retries is not None
                        else 0
                    ),
                    priority=(
                        task_config.priority if task_config.priority is not None else 5
                    ),
                    timeout_seconds=task_config.timeout_seconds,
                    idempotency_key=None,
                )
                # Convert task request to SQL command string
                insert_task_command = self._create_insert_task_command(task_request)
                create_cron_job_for_task_sync(
                    conn,
                    create_cron_job_for_task.CreateCronJobForTaskParams(
                        schedule=cron,
                        command=insert_task_command,
                        jobname=cron_job_name,
                    ),
                )
            else:
                turn_off_cron_for_task_sync(
                    conn,
                    turn_off_cron_for_task.TurnOffCronForTaskParams(
                        jobname=cron_job_name
                    ),
                )

    def _create_insert_task_command(self, task: EnqueueTaskRequest) -> str:
        """Create SQL command for inserting a task (used by cron jobs)."""
        # This is a simplified version - you may need to adjust based on your exact SQL
        return f"""
        INSERT INTO hyrex_task_run (
            id, durable_id, root_id, parent_id, task_name, args, queue,
            max_retries, priority, timeout_seconds, idempotency_key, status,
            workflow_run_id, workflow_dependencies, attempt_number, queued
        ) VALUES (
            '{task.id}'::uuid, '{task.durable_id}'::uuid, '{task.root_id}'::uuid,
            {f"'{task.parent_id}'::uuid" if task.parent_id else 'NULL'},
            '{task.task_name}', '{json.dumps(task.args)}'::json, '{task.queue}',
            {task.max_retries}, {task.priority},
            {task.timeout_seconds if task.timeout_seconds else 'NULL'},
            {f"'{task.idempotency_key}'" if task.idempotency_key else 'NULL'},
            '{task.status.value}'::task_run_status,
            {f"'{task.workflow_run_id}'::uuid" if task.workflow_run_id else 'NULL'},
            {f"ARRAY{task.workflow_dependencies}::uuid[]" if task.workflow_dependencies else 'NULL'},
            0, CURRENT_TIMESTAMP
        )
        """

    def register_workflow(
        self,
        name: str,
        source_code: str,
        workflow_dag_json: dict,
        workflow_arg_schema: Type[BaseModel] | None,
        default_config: dict,
        cron: str | None,
    ):
        with self.transaction() as conn:
            register_workflow_sync(
                conn,
                register_workflow.RegisterWorkflowParams(
                    workflow_name=name,
                    cron_expr=cron,
                    source_code=source_code,
                    dag_structure=json.dumps(workflow_dag_json),
                ),
            )

    def send_workflow_run(self, workflow_run_request: WorkflowRunRequest) -> UUID:
        with self.transaction() as conn:
            workflow_run_id = create_workflow_run_sync(
                conn,
                create_workflow_run.CreateWorkflowRunParams(
                    workflow_run_id=workflow_run_request.id,
                    workflow_name=workflow_run_request.workflow_name,
                    args=json.dumps(
                        workflow_run_request.args, default=pydantic_aware_default
                    ),
                    queue=workflow_run_request.queue,
                    timeout_seconds=workflow_run_request.timeout_seconds,
                    idempotency_key=workflow_run_request.idempotency_key,
                ),
            )
            if workflow_run_id is None:
                raise ValueError("Insert workflow run failed.")
            return workflow_run_id

    def advance_workflow_run(self, workflow_run_id: UUID):
        self.logger.info(f"Advancing workflow run {workflow_run_id}")

        with self.transaction() as conn:
            # First update workflow run status based on task runs
            updated_row = set_workflow_run_status_based_on_task_runs_sync(
                conn,
                set_workflow_run_status_based_on_task_runs.SetWorkflowRunStatusBasedOnTaskRunsParams(
                    workflow_run_id=workflow_run_id
                ),
            )

            if updated_row and updated_row.status in ("FAILED", "SUCCESS"):
                return None

            # Advance the workflow
            list(
                advance_workflow_run_func_sync(
                    conn,
                    advance_workflow_run_func.AdvanceWorkflowRunFuncParams(
                        workflow_run_id=workflow_run_id
                    ),
                )
            )
            return None

    def get_workflow_run_args(self, workflow_run_id: UUID) -> dict:
        with self.transaction() as conn:
            workflow_run = get_workflow_run_by_id_sync(
                conn,
                get_workflow_run_by_id.GetWorkflowRunByIdParams(id=workflow_run_id),
            )
            return workflow_run.args if workflow_run else None

    def acquire_scheduler_lock(self, worker_name: str) -> int | None:
        with self.transaction() as conn:
            lock_id = acquire_scheduler_lock_sync(
                conn,
                acquire_scheduler_lock.AcquireSchedulerLockParams(
                    worker_name=worker_name, duration=timedelta(minutes=2)
                ),
            )
            return lock_id

    def pull_cron_job_expressions(self) -> list[CronJob]:
        with self.transaction() as conn:
            rows = list(
                pull_active_cron_expressions_sync(
                    conn, pull_active_cron_expressions.PullActiveCronExpressionsParams()
                )
            )
            return [
                CronJob(
                    jobid=row.jobid,
                    schedule=row.schedule,
                    command=row.command,
                    active=row.active,
                    jobname=row.jobname,
                    activated_at=row.activated_at,
                    scheduled_jobs_confirmed_until=row.scheduled_jobs_confirmed_until,
                    should_backfill=row.should_backfill,
                )
                for row in rows
            ]

    def update_cron_job_confirmation_timestamp(self, jobid: int):
        with self.transaction() as conn:
            update_cron_job_confirmation_ts_sync(
                conn,
                update_cron_job_confirmation_ts.UpdateCronJobConfirmationTsParams(
                    jobid=jobid
                ),
            )

    def schedule_cron_job_runs(self, cron_job_runs: List[CronJobRun]) -> None:
        if not cron_job_runs:
            return

        # Check all jobs have same ID
        all_same_id = all(job.jobid == cron_job_runs[0].jobid for job in cron_job_runs)
        if not all_same_id:
            job_ids = [job.jobid for job in cron_job_runs]
            self.logger.error(f"Got jobIds {job_ids}, {cron_job_runs}")
            raise ValueError(
                "All cronJobsRuns submitted here need to have the same job id."
            )

        # Execute the SQL
        with self.transaction() as conn:
            # Convert cron job runs to JSON format expected by SQLC
            cron_jobs_json = [
                {
                    "jobid": job.jobid,
                    "command": job.command,
                    "schedule_time": job.schedule_time.isoformat(),
                }
                for job in cron_job_runs
            ]

            schedule_cron_job_runs_json_sync(
                conn,
                schedule_cron_job_runs_json.ScheduleCronJobRunsJsonParams(
                    runs_json=json.dumps(cron_jobs_json)
                ),
            )

        # Update confirmation timestamp
        self.update_cron_job_confirmation_timestamp(cron_job_runs[0].jobid)

    def execute_queued_cron_job_run(self) -> str | None:
        with self.transaction() as conn:
            result = trigger_execute_queued_cron_job_sync(
                conn,
                trigger_execute_queued_cron_job.TriggerExecuteQueuedCronJobParams(),
            )
            return result  # "executed" or "not_found"

    def register_cron_sql_query(
        self,
        cron_job_name: str,
        cron_sql_query: str,
        cron_expr: str,
        should_backfill: bool,
    ) -> None:
        with self.transaction() as conn:
            create_cron_job_for_sql_query_sync(
                conn,
                create_cron_job_for_sql_query.CreateCronJobForSqlQueryParams(
                    schedule=cron_expr,
                    command=cron_sql_query,
                    jobname=cron_job_name,
                    should_backfill=should_backfill,
                ),
            )

    def release_scheduler_lock(self, worker_name: str) -> None:
        with self.transaction() as conn:
            release_scheduler_lock_sync(
                conn,
                release_scheduler_lock.ReleaseSchedulerLockParams(
                    worker_name=worker_name
                ),
            )

    def get_durable_run_tasks(self, durable_id: UUID) -> list[TaskRun]:
        with self.transaction() as conn:
            results = list(
                get_task_attempts_by_durable_id_sync(
                    conn,
                    get_task_attempts_by_durable_id.GetTaskAttemptsByDurableIdParams(
                        durable_id=durable_id
                    ),
                )
            )

            task_runs = []
            for row in results:
                # Fetch result if exists
                result_data = {}
                if row.finished and row.status == "SUCCESS":
                    result_row = fetch_result_sync(
                        conn, fetch_result.FetchResultParams(task_id=row.id)
                    )
                    if result_row:
                        result_data = result_row.result

                task_run = TaskRun(
                    id=row.id,
                    task_name=row.task_name,
                    max_retries=row.max_retries,
                    attempt_number=row.attempt_number,
                    status=row.status,
                    queued=row.queued,
                    started=row.started,
                    finished=row.finished,
                    result=result_data,
                )
                task_runs.append(task_run)

            return task_runs

    def get_workflow_durable_runs(self, workflow_run_id: UUID) -> list[UUID]:
        with self.transaction() as conn:
            task_runs = list(
                get_workflow_run_task_runs_sync(
                    conn,
                    get_workflow_run_task_runs.GetWorkflowRunTaskRunsParams(
                        workflow_run_id=workflow_run_id
                    ),
                )
            )
            # Get unique durable IDs
            durable_ids = list(set(run.durable_id for run in task_runs))
            return durable_ids

    def try_to_cancel_durable_run(self, durable_id: UUID):
        # This would need to be implemented based on your business logic
        # For now, we'll just cancel all tasks with this durable_id
        with self.transaction() as conn:
            # You might need to add a specific SQLC query for this
            # For now, using raw SQL as an example
            conn.execute(
                sqlalchemy.text(
                    "UPDATE hyrex_task_run SET status = 'CANCEL_REQUESTED'::task_run_status "
                    "WHERE durable_id = :durable_id AND status IN ('QUEUED', 'RUNNING')"
                ),
                {"durable_id": durable_id},
            )

    def update_executor_queues(self, executor_id: UUID, queues: list[str]):
        with self.transaction() as conn:
            update_queues_on_executor_sync(
                conn,
                update_queues_on_executor.UpdateQueuesOnExecutorParams(
                    id=executor_id, queues=queues
                ),
            )

    def get_result(self, task_id: UUID) -> dict:
        with self.transaction() as conn:
            result = fetch_result_sync(
                conn, fetch_result.FetchResultParams(task_id=task_id)
            )
            return result.result if result else None

    def kv_set(self, key: str, value: str) -> None:
        with self.transaction() as conn:
            set_value_sync(
                conn,
                kv_set_value.SetValueParams(key=key, value=value)
            )

    def kv_get(self, key: str) -> str | None:
        with self.transaction() as conn:
            return get_value_sync(
                conn,
                kv_get_value.GetValueParams(key=key)
            )

    def kv_delete(self, key: str) -> None:
        with self.transaction() as conn:
            delete_value_sync(
                conn,
                kv_delete_value.DeleteValueParams(key=key)
            )



from datetime import datetime
from enum import StrEnum
from uuid import UUID

from pydantic import BaseModel


class QueuePattern(BaseModel):
    glob_pattern: str
    postgres_pattern: str


class TaskStatus(StrEnum):
    success = "SUCCESS"
    failed = "FAILED"
    running = "RUNNING"
    queued = "QUEUED"
    up_for_cancel = "UP_FOR_CANCEL"
    canceled = "CANCELED"
    lost = "LOST"
    stopped = "STOPPED"
    skipped = "SKIPPED"
    await_deps = "AWAIT_DEPS"
    await_start_time = "AWAIT_START_TIME"


class EnqueueTaskRequest(BaseModel):
    id: UUID
    durable_id: UUID
    root_id: UUID
    parent_id: UUID | None
    task_name: str
    args: dict
    queue: str
    max_retries: int
    priority: int
    timeout_seconds: int | None
    idempotency_key: str | None
    status: TaskStatus
    workflow_run_id: UUID | None
    workflow_dependencies: list[UUID] | None


class DequeuedTask(BaseModel):
    id: UUID
    durable_id: UUID
    root_id: UUID
    parent_id: UUID | None
    task_name: str
    args: dict
    queue: str
    priority: int
    timeout_seconds: int | None
    scheduled_start: datetime | None
    queued: datetime
    started: datetime
    workflow_run_id: UUID | None
    attempt_number: int
    max_retries: int


# For tracking durable runs:
class TaskRun(BaseModel):
    id: UUID
    task_name: str
    max_retries: int
    attempt_number: int
    status: TaskStatus
    queued: datetime
    started: datetime | None
    finished: datetime | None
    result: dict | None


class WorkflowStatus(StrEnum):
    success = "SUCCESS"
    failed = "FAILED"
    running = "RUNNING"
    up_for_cancel = "UP_FOR_CANCEL"
    canceled = "CANCELED"
    asleep = "ASLEEP"


class WorkflowRunRequest(BaseModel):
    id: UUID
    workflow_name: str
    args: dict
    queue: str
    timeout_seconds: int | None
    idempotency_key: str | None


class CronJob(BaseModel):
    jobid: int
    schedule: str
    command: str
    active: bool
    jobname: str
    activated_at: datetime
    scheduled_jobs_confirmed_until: datetime
    should_backfill: bool


class CronJobRun(BaseModel):
    jobid: int
    command: str
    schedule_time: datetime



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: create_schedule_cron_job_runs_func.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_SCHEDULE_CRON_JOB_RUNS_FUNC = """-- name: create_schedule_cron_job_runs_func \\:exec
DO $$
BEGIN
    -- First create the composite type for input if it doesn't exist
    IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'cron_job_run_input') THEN
        CREATE TYPE cron_job_run_input AS (
            jobid BIGINT,
            command TEXT,
            schedule_time TIMESTAMPTZ
        );
    END IF;

    -- Create the function to schedule cron job runs
    CREATE OR REPLACE FUNCTION schedule_cron_job_runs(
        p_runs cron_job_run_input[]
    ) RETURNS TABLE (
        success BOOLEAN,
        message TEXT,
        inserted_count INT,
        runids BIGINT[]
    ) AS $func$
    DECLARE
        v_jobid BIGINT;
        v_run cron_job_run_input;
        v_inserted_runids BIGINT[] \\:= '{}';
        v_inserted_count INT \\:= 0;
        v_runid BIGINT;
    BEGIN
        -- Validate input
        IF p_runs IS NULL OR array_length(p_runs, 1) IS NULL THEN
            RETURN QUERY SELECT FALSE, 'No cron job runs provided', 0, NULL\\:\\:BIGINT[];
            RETURN;
        END IF;
        
        -- Extract the first jobid for validation
        v_jobid \\:= (p_runs[1]).jobid;
        
        -- Validate all runs have the same jobid
        FOREACH v_run IN ARRAY p_runs
        LOOP
            IF v_run.jobid != v_jobid THEN
                RETURN QUERY SELECT FALSE, 'All cron job runs must have the same job id', 0, NULL\\:\\:BIGINT[];
                RETURN;
            END IF;
        END LOOP;
        
        -- Insert all runs
        FOREACH v_run IN ARRAY p_runs
        LOOP
            INSERT INTO hyrex_cron_job_run_details (
                jobid,
                command,
                status,
                schedule_time
            )
            VALUES (
                v_run.jobid,
                v_run.command,
                'QUEUED'\\:\\:cron_job_status_enum,
                v_run.schedule_time
            )
            ON CONFLICT (jobid, schedule_time) DO NOTHING
            RETURNING runid INTO v_runid;
            
            -- If a row was inserted (no conflict), add to our array
            IF v_runid IS NOT NULL THEN
                v_inserted_runids \\:= array_append(v_inserted_runids, v_runid);
                v_inserted_count \\:= v_inserted_count + 1;
                v_runid \\:= NULL; -- Reset for next iteration
            END IF;
        END LOOP;
        
        -- Update the cron job confirmation timestamp
        UPDATE hyrex_cron_job
        SET scheduled_jobs_confirmed_until = now()
        WHERE jobid = v_jobid;
        
        -- Log the scheduling event
        INSERT INTO hyrex_system_logs (id, timestamp, event_name, event_body)
        VALUES (
            gen_random_uuid(),
            CURRENT_TIMESTAMP,
            'cron_jobs_scheduled',
            json_build_object(
                'jobid', v_jobid,
                'scheduled_count', array_length(p_runs, 1),
                'inserted_count', v_inserted_count,
                'runids', v_inserted_runids
            )
        );
        
        RETURN QUERY SELECT 
            TRUE, 
            format('Scheduled %s cron job runs, %s inserted', array_length(p_runs, 1), v_inserted_count),
            v_inserted_count,
            v_inserted_runids;
            
    EXCEPTION
        WHEN OTHERS THEN
            -- Log the error
            INSERT INTO hyrex_system_logs (id, timestamp, event_name, event_body)
            VALUES (
                gen_random_uuid(),
                CURRENT_TIMESTAMP,
                'cron_jobs_schedule_error',
                json_build_object(
                    'jobid', v_jobid,
                    'error', SQLERRM,
                    'sqlstate', SQLSTATE
                )
            );
            
            RETURN QUERY SELECT FALSE, format('Error scheduling cron jobs\\: %s', SQLERRM), 0, NULL\\:\\:BIGINT[];
    END;
    $func$ LANGUAGE plpgsql;
END
$$
"""


@dataclasses.dataclass()
class CreateScheduleCronJobRunsFuncParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_schedule_cron_job_runs_func(self, arg: CreateScheduleCronJobRunsFuncParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_SCHEDULE_CRON_JOB_RUNS_FUNC))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_schedule_cron_job_runs_func(self, arg: CreateScheduleCronJobRunsFuncParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_SCHEDULE_CRON_JOB_RUNS_FUNC))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: count_queued_hype_cron_job_runs.sql
import dataclasses
from typing import Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


COUNT_QUEUED_HYPE_CRON_JOB_RUNS = """-- name: count_queued_hype_cron_job_runs \\:one
  SELECT COUNT(*)\\:\\:int AS count
  FROM hype_cron_job_run_details
  WHERE status = 'QUEUED'
    AND schedule_time <= NOW()
"""


@dataclasses.dataclass()
class CountQueuedHypeCronJobRunsParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def count_queued_hype_cron_job_runs(self, arg: CountQueuedHypeCronJobRunsParams) -> Optional[int]:
        row = self._conn.execute(sqlalchemy.text(COUNT_QUEUED_HYPE_CRON_JOB_RUNS)).first()
        if row is None:
            return None
        return row[0]


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def count_queued_hype_cron_job_runs(self, arg: CountQueuedHypeCronJobRunsParams) -> Optional[int]:
        row = (await self._conn.execute(sqlalchemy.text(COUNT_QUEUED_HYPE_CRON_JOB_RUNS))).first()
        if row is None:
            return None
        return row[0]



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: _07_create_hyrex_task_def_table.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_TASK_DEF_TABLE = """-- name: create_task_def_table \\:exec
CREATE TABLE IF NOT EXISTS hyrex_task_def
(
    task_name        TEXT NOT NULL PRIMARY KEY,
    cron_expr        TEXT,
    source_code      TEXT,
    arg_schema       JSON,
    queue            TEXT,
    priority         INTEGER,
    max_retries      INTEGER,
    timeout_seconds  INTEGER CHECK (timeout_seconds IS NULL OR timeout_seconds > 0),
    last_updated     TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP
)
"""


@dataclasses.dataclass()
class CreateTaskDefTableParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_task_def_table(self, arg: CreateTaskDefTableParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_TASK_DEF_TABLE))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_task_def_table(self, arg: CreateTaskDefTableParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_TASK_DEF_TABLE))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: update_executor_stats.sql
import dataclasses
import datetime
from typing import Any, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


UPDATE_EXECUTOR_STATS = """-- name: update_executor_stats \\:one
WITH updated AS (
    UPDATE hyrex_executor
        SET last_heartbeat = CURRENT_TIMESTAMP,
            stats = :p1
        WHERE hyrex_executor.id = :p2
            AND hyrex_executor.status = 'RUNNING'\\:\\:executor_status
        RETURNING id, status, last_heartbeat, stats),
     executor_state AS (SELECT CASE WHEN u.id IS NOT NULL THEN 'ACCEPTED' ELSE 'REJECTED' END AS result,
                               COALESCE(u.status, e.status)\\:\\:text                             AS status,
                               COALESCE(u.last_heartbeat, e.last_heartbeat)                   AS last_heartbeat,
                               COALESCE(u.stats, e.stats)                                     AS stats
                        FROM hyrex_executor e
                                 LEFT JOIN updated u ON e.id = u.id
                        WHERE e.id = :p2),
     insert_log AS (
         INSERT INTO hyrex_system_logs (id, timestamp, event_name, event_body)
             SELECT gen_random_uuid(),
                    CURRENT_TIMESTAMP,
                    'HEARTBEAT_REJECTED',
                    json_build_object('executor_id', :p2, 'current_status', es.status, 'stats', :p1)
             FROM executor_state es
             WHERE es.result = 'REJECTED'
             RETURNING NULL)
SELECT es.result, es.status, es.last_heartbeat, es.stats
FROM executor_state es
"""


@dataclasses.dataclass()
class UpdateExecutorStatsParams:
    stats: Optional[Any]
    id: uuid.UUID


@dataclasses.dataclass()
class UpdateExecutorStatsRow:
    result: str
    status: str
    last_heartbeat: Optional[datetime.datetime]
    stats: Optional[Any]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def update_executor_stats(self, arg: UpdateExecutorStatsParams) -> Optional[UpdateExecutorStatsRow]:
        row = self._conn.execute(sqlalchemy.text(UPDATE_EXECUTOR_STATS), {"p1": arg.stats, "p2": arg.id}).first()
        if row is None:
            return None
        return UpdateExecutorStatsRow(
            result=row[0],
            status=row[1],
            last_heartbeat=row[2],
            stats=row[3],
        )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def update_executor_stats(self, arg: UpdateExecutorStatsParams) -> Optional[UpdateExecutorStatsRow]:
        row = (await self._conn.execute(sqlalchemy.text(UPDATE_EXECUTOR_STATS), {"p1": arg.stats, "p2": arg.id})).first()
        if row is None:
            return None
        return UpdateExecutorStatsRow(
            result=row[0],
            status=row[1],
            last_heartbeat=row[2],
            stats=row[3],
        )



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: disconnect_executor.sql
import dataclasses
from typing import Any, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


DISCONNECT_EXECUTOR = """-- name: disconnect_executor \\:exec
UPDATE hyrex_executor
SET stopped        = CURRENT_TIMESTAMP,
    last_heartbeat = CURRENT_TIMESTAMP,
    stats          = :p1,
    status         = 'SHUTDOWN'\\:\\:executor_status
WHERE id = :p2
"""


@dataclasses.dataclass()
class DisconnectExecutorParams:
    stats: Optional[Any]
    id: uuid.UUID


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def disconnect_executor(self, arg: DisconnectExecutorParams) -> None:
        self._conn.execute(sqlalchemy.text(DISCONNECT_EXECUTOR), {"p1": arg.stats, "p2": arg.id})


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def disconnect_executor(self, arg: DisconnectExecutorParams) -> None:
        await self._conn.execute(sqlalchemy.text(DISCONNECT_EXECUTOR), {"p1": arg.stats, "p2": arg.id})



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: turn_off_cron_for_task.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


TURN_OFF_CRON_FOR_TASK = """-- name: turn_off_cron_for_task \\:exec
UPDATE hyrex_cron_job
SET active   = false,
    schedule = NULL
WHERE jobname = :p1
  AND job_source = 'TASK'
"""


@dataclasses.dataclass()
class TurnOffCronForTaskParams:
    jobname: str


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def turn_off_cron_for_task(self, arg: TurnOffCronForTaskParams) -> None:
        self._conn.execute(sqlalchemy.text(TURN_OFF_CRON_FOR_TASK), {"p1": arg.jobname})


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def turn_off_cron_for_task(self, arg: TurnOffCronForTaskParams) -> None:
        await self._conn.execute(sqlalchemy.text(TURN_OFF_CRON_FOR_TASK), {"p1": arg.jobname})



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import requests_pb2 as requests__pb2

GRPC_GENERATED_VERSION = '1.71.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in gateway_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class GatewayServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.TestConnection = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/TestConnection',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.WriteLogs = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/WriteLogs',
                request_serializer=requests__pb2.WriteLogsRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.Enqueue = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/Enqueue',
                request_serializer=requests__pb2.EnqueueRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.Dequeue = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/Dequeue',
                request_serializer=requests__pb2.DequeueRequest.SerializeToString,
                response_deserializer=requests__pb2.DequeueResponse.FromString,
                _registered_method=True)
        self.GetQueues = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetQueues',
                request_serializer=requests__pb2.GetQueuesRequest.SerializeToString,
                response_deserializer=requests__pb2.GetQueuesResponse.FromString,
                _registered_method=True)
        self.GetTaskRunStatus = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetTaskRunStatus',
                request_serializer=requests__pb2.GetTaskRunStatusRequest.SerializeToString,
                response_deserializer=requests__pb2.GetTaskRunStatusResponse.FromString,
                _registered_method=True)
        self.MarkSuccess = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/MarkSuccess',
                request_serializer=requests__pb2.MarkSuccessRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.MarkFailed = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/MarkFailed',
                request_serializer=requests__pb2.MarkFailedRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.GetTaskRun = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetTaskRun',
                request_serializer=requests__pb2.GetTaskRunRequest.SerializeToString,
                response_deserializer=requests__pb2.GetTaskRunResponse.FromString,
                _registered_method=True)
        self.GetDurableTaskRuns = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetDurableTaskRuns',
                request_serializer=requests__pb2.GetDurableTaskRunsRequest.SerializeToString,
                response_deserializer=requests__pb2.GetDurableTaskRunsResponse.FromString,
                _registered_method=True)
        self.RetryTaskRun = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/RetryTaskRun',
                request_serializer=requests__pb2.RetryTaskRunRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.TaskRunHeartbeat = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/TaskRunHeartbeat',
                request_serializer=requests__pb2.TaskRunHeartbeatRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.TryToCancelDurableRun = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/TryToCancelDurableRun',
                request_serializer=requests__pb2.TryToCancelDurableRunRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.GetTaskRunsUpForCancel = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetTaskRunsUpForCancel',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=requests__pb2.GetTaskRunsUpForCancelResponse.FromString,
                _registered_method=True)
        self.MarkCanceled = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/MarkCanceled',
                request_serializer=requests__pb2.MarkCanceledRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.KVStoreSet = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/KVStoreSet',
                request_serializer=requests__pb2.KVStoreSetRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.KVStoreGet = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/KVStoreGet',
                request_serializer=requests__pb2.KVStoreGetRequest.SerializeToString,
                response_deserializer=requests__pb2.KVStoreGetResponse.FromString,
                _registered_method=True)
        self.KVStoreDelete = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/KVStoreDelete',
                request_serializer=requests__pb2.KVStoreDeleteRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.MarkRunningTasksLost = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/MarkRunningTasksLost',
                request_serializer=requests__pb2.MarkRunningTasksLostRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.RegisterTaskDef = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/RegisterTaskDef',
                request_serializer=requests__pb2.RegisterTaskDefRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.GetTaskDef = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetTaskDef',
                request_serializer=requests__pb2.GetTaskDefRequest.SerializeToString,
                response_deserializer=requests__pb2.GetTaskDefResponse.FromString,
                _registered_method=True)
        self.GetAllTaskDefs = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetAllTaskDefs',
                request_serializer=requests__pb2.GetAllTaskDefsRequest.SerializeToString,
                response_deserializer=requests__pb2.GetAllTaskDefsResponse.FromString,
                _registered_method=True)
        self.RegisterExecutor = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/RegisterExecutor',
                request_serializer=requests__pb2.RegisterExecutorRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.ExecutorHeartbeat = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/ExecutorHeartbeat',
                request_serializer=requests__pb2.ExecutorHeartbeatRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.UpdateExecutorStats = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/UpdateExecutorStats',
                request_serializer=requests__pb2.UpdateExecutorStatsRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.UpdateExecutorQueues = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/UpdateExecutorQueues',
                request_serializer=requests__pb2.UpdateExecutorQueuesRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.DisconnectExecutor = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/DisconnectExecutor',
                request_serializer=requests__pb2.DisconnectExecutorRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.RegisterApp = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/RegisterApp',
                request_serializer=requests__pb2.RegisterAppRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.SetLogLink = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/SetLogLink',
                request_serializer=requests__pb2.SetLogLinkRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.RegisterWorkflow = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/RegisterWorkflow',
                request_serializer=requests__pb2.RegisterWorkflowRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.SendWorkflowRun = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/SendWorkflowRun',
                request_serializer=requests__pb2.SendWorkflowRunRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.GetWorkflowRunArgs = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetWorkflowRunArgs',
                request_serializer=requests__pb2.GetWorkflowRunArgsRequest.SerializeToString,
                response_deserializer=requests__pb2.GetWorkflowRunArgsResponse.FromString,
                _registered_method=True)
        self.AdvanceWorkflowRun = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/AdvanceWorkflowRun',
                request_serializer=requests__pb2.AdvanceWorkflowRunRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.GetWorkflowDurableRuns = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetWorkflowDurableRuns',
                request_serializer=requests__pb2.GetWorkflowDurableRunsRequest.SerializeToString,
                response_deserializer=requests__pb2.GetWorkflowDurableRunsResponse.FromString,
                _registered_method=True)


class GatewayServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def TestConnection(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def WriteLogs(self, request, context):
        """---- S3 requests ----
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Enqueue(self, request, context):
        """---- Valkey requests ----
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Dequeue(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetQueues(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTaskRunStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def MarkSuccess(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def MarkFailed(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTaskRun(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetDurableTaskRuns(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RetryTaskRun(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TaskRunHeartbeat(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TryToCancelDurableRun(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTaskRunsUpForCancel(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def MarkCanceled(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def KVStoreSet(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def KVStoreGet(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def KVStoreDelete(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def MarkRunningTasksLost(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RegisterTaskDef(self, request, context):
        """---- Postgres requests ----
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTaskDef(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAllTaskDefs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RegisterExecutor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExecutorHeartbeat(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateExecutorStats(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateExecutorQueues(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DisconnectExecutor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RegisterApp(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetLogLink(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RegisterWorkflow(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendWorkflowRun(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetWorkflowRunArgs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AdvanceWorkflowRun(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetWorkflowDurableRuns(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GatewayServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'TestConnection': grpc.unary_unary_rpc_method_handler(
                    servicer.TestConnection,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'WriteLogs': grpc.unary_unary_rpc_method_handler(
                    servicer.WriteLogs,
                    request_deserializer=requests__pb2.WriteLogsRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'Enqueue': grpc.unary_unary_rpc_method_handler(
                    servicer.Enqueue,
                    request_deserializer=requests__pb2.EnqueueRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'Dequeue': grpc.unary_unary_rpc_method_handler(
                    servicer.Dequeue,
                    request_deserializer=requests__pb2.DequeueRequest.FromString,
                    response_serializer=requests__pb2.DequeueResponse.SerializeToString,
            ),
            'GetQueues': grpc.unary_unary_rpc_method_handler(
                    servicer.GetQueues,
                    request_deserializer=requests__pb2.GetQueuesRequest.FromString,
                    response_serializer=requests__pb2.GetQueuesResponse.SerializeToString,
            ),
            'GetTaskRunStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTaskRunStatus,
                    request_deserializer=requests__pb2.GetTaskRunStatusRequest.FromString,
                    response_serializer=requests__pb2.GetTaskRunStatusResponse.SerializeToString,
            ),
            'MarkSuccess': grpc.unary_unary_rpc_method_handler(
                    servicer.MarkSuccess,
                    request_deserializer=requests__pb2.MarkSuccessRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'MarkFailed': grpc.unary_unary_rpc_method_handler(
                    servicer.MarkFailed,
                    request_deserializer=requests__pb2.MarkFailedRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'GetTaskRun': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTaskRun,
                    request_deserializer=requests__pb2.GetTaskRunRequest.FromString,
                    response_serializer=requests__pb2.GetTaskRunResponse.SerializeToString,
            ),
            'GetDurableTaskRuns': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDurableTaskRuns,
                    request_deserializer=requests__pb2.GetDurableTaskRunsRequest.FromString,
                    response_serializer=requests__pb2.GetDurableTaskRunsResponse.SerializeToString,
            ),
            'RetryTaskRun': grpc.unary_unary_rpc_method_handler(
                    servicer.RetryTaskRun,
                    request_deserializer=requests__pb2.RetryTaskRunRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'TaskRunHeartbeat': grpc.unary_unary_rpc_method_handler(
                    servicer.TaskRunHeartbeat,
                    request_deserializer=requests__pb2.TaskRunHeartbeatRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'TryToCancelDurableRun': grpc.unary_unary_rpc_method_handler(
                    servicer.TryToCancelDurableRun,
                    request_deserializer=requests__pb2.TryToCancelDurableRunRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'GetTaskRunsUpForCancel': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTaskRunsUpForCancel,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=requests__pb2.GetTaskRunsUpForCancelResponse.SerializeToString,
            ),
            'MarkCanceled': grpc.unary_unary_rpc_method_handler(
                    servicer.MarkCanceled,
                    request_deserializer=requests__pb2.MarkCanceledRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'KVStoreSet': grpc.unary_unary_rpc_method_handler(
                    servicer.KVStoreSet,
                    request_deserializer=requests__pb2.KVStoreSetRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'KVStoreGet': grpc.unary_unary_rpc_method_handler(
                    servicer.KVStoreGet,
                    request_deserializer=requests__pb2.KVStoreGetRequest.FromString,
                    response_serializer=requests__pb2.KVStoreGetResponse.SerializeToString,
            ),
            'KVStoreDelete': grpc.unary_unary_rpc_method_handler(
                    servicer.KVStoreDelete,
                    request_deserializer=requests__pb2.KVStoreDeleteRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'MarkRunningTasksLost': grpc.unary_unary_rpc_method_handler(
                    servicer.MarkRunningTasksLost,
                    request_deserializer=requests__pb2.MarkRunningTasksLostRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'RegisterTaskDef': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterTaskDef,
                    request_deserializer=requests__pb2.RegisterTaskDefRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'GetTaskDef': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTaskDef,
                    request_deserializer=requests__pb2.GetTaskDefRequest.FromString,
                    response_serializer=requests__pb2.GetTaskDefResponse.SerializeToString,
            ),
            'GetAllTaskDefs': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAllTaskDefs,
                    request_deserializer=requests__pb2.GetAllTaskDefsRequest.FromString,
                    response_serializer=requests__pb2.GetAllTaskDefsResponse.SerializeToString,
            ),
            'RegisterExecutor': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterExecutor,
                    request_deserializer=requests__pb2.RegisterExecutorRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'ExecutorHeartbeat': grpc.unary_unary_rpc_method_handler(
                    servicer.ExecutorHeartbeat,
                    request_deserializer=requests__pb2.ExecutorHeartbeatRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'UpdateExecutorStats': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateExecutorStats,
                    request_deserializer=requests__pb2.UpdateExecutorStatsRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'UpdateExecutorQueues': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateExecutorQueues,
                    request_deserializer=requests__pb2.UpdateExecutorQueuesRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'DisconnectExecutor': grpc.unary_unary_rpc_method_handler(
                    servicer.DisconnectExecutor,
                    request_deserializer=requests__pb2.DisconnectExecutorRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'RegisterApp': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterApp,
                    request_deserializer=requests__pb2.RegisterAppRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'SetLogLink': grpc.unary_unary_rpc_method_handler(
                    servicer.SetLogLink,
                    request_deserializer=requests__pb2.SetLogLinkRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'RegisterWorkflow': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterWorkflow,
                    request_deserializer=requests__pb2.RegisterWorkflowRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'SendWorkflowRun': grpc.unary_unary_rpc_method_handler(
                    servicer.SendWorkflowRun,
                    request_deserializer=requests__pb2.SendWorkflowRunRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'GetWorkflowRunArgs': grpc.unary_unary_rpc_method_handler(
                    servicer.GetWorkflowRunArgs,
                    request_deserializer=requests__pb2.GetWorkflowRunArgsRequest.FromString,
                    response_serializer=requests__pb2.GetWorkflowRunArgsResponse.SerializeToString,
            ),
            'AdvanceWorkflowRun': grpc.unary_unary_rpc_method_handler(
                    servicer.AdvanceWorkflowRun,
                    request_deserializer=requests__pb2.AdvanceWorkflowRunRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'GetWorkflowDurableRuns': grpc.unary_unary_rpc_method_handler(
                    servicer.GetWorkflowDurableRuns,
                    request_deserializer=requests__pb2.GetWorkflowDurableRunsRequest.FromString,
                    response_serializer=requests__pb2.GetWorkflowDurableRunsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'hyrex.performanceserver.GatewayService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('hyrex.performanceserver.GatewayService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class GatewayService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def TestConnection(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/TestConnection',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def WriteLogs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/WriteLogs',
            requests__pb2.WriteLogsRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Enqueue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/Enqueue',
            requests__pb2.EnqueueRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Dequeue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/Dequeue',
            requests__pb2.DequeueRequest.SerializeToString,
            requests__pb2.DequeueResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetQueues(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetQueues',
            requests__pb2.GetQueuesRequest.SerializeToString,
            requests__pb2.GetQueuesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTaskRunStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetTaskRunStatus',
            requests__pb2.GetTaskRunStatusRequest.SerializeToString,
            requests__pb2.GetTaskRunStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def MarkSuccess(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/MarkSuccess',
            requests__pb2.MarkSuccessRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def MarkFailed(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/MarkFailed',
            requests__pb2.MarkFailedRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTaskRun(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetTaskRun',
            requests__pb2.GetTaskRunRequest.SerializeToString,
            requests__pb2.GetTaskRunResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetDurableTaskRuns(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetDurableTaskRuns',
            requests__pb2.GetDurableTaskRunsRequest.SerializeToString,
            requests__pb2.GetDurableTaskRunsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RetryTaskRun(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/RetryTaskRun',
            requests__pb2.RetryTaskRunRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def TaskRunHeartbeat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/TaskRunHeartbeat',
            requests__pb2.TaskRunHeartbeatRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def TryToCancelDurableRun(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/TryToCancelDurableRun',
            requests__pb2.TryToCancelDurableRunRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTaskRunsUpForCancel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetTaskRunsUpForCancel',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            requests__pb2.GetTaskRunsUpForCancelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def MarkCanceled(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/MarkCanceled',
            requests__pb2.MarkCanceledRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def KVStoreSet(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/KVStoreSet',
            requests__pb2.KVStoreSetRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def KVStoreGet(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/KVStoreGet',
            requests__pb2.KVStoreGetRequest.SerializeToString,
            requests__pb2.KVStoreGetResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def KVStoreDelete(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/KVStoreDelete',
            requests__pb2.KVStoreDeleteRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def MarkRunningTasksLost(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/MarkRunningTasksLost',
            requests__pb2.MarkRunningTasksLostRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RegisterTaskDef(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/RegisterTaskDef',
            requests__pb2.RegisterTaskDefRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTaskDef(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetTaskDef',
            requests__pb2.GetTaskDefRequest.SerializeToString,
            requests__pb2.GetTaskDefResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetAllTaskDefs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetAllTaskDefs',
            requests__pb2.GetAllTaskDefsRequest.SerializeToString,
            requests__pb2.GetAllTaskDefsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RegisterExecutor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/RegisterExecutor',
            requests__pb2.RegisterExecutorRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ExecutorHeartbeat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/ExecutorHeartbeat',
            requests__pb2.ExecutorHeartbeatRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UpdateExecutorStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/UpdateExecutorStats',
            requests__pb2.UpdateExecutorStatsRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UpdateExecutorQueues(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/UpdateExecutorQueues',
            requests__pb2.UpdateExecutorQueuesRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DisconnectExecutor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/DisconnectExecutor',
            requests__pb2.DisconnectExecutorRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RegisterApp(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/RegisterApp',
            requests__pb2.RegisterAppRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SetLogLink(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/SetLogLink',
            requests__pb2.SetLogLinkRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RegisterWorkflow(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/RegisterWorkflow',
            requests__pb2.RegisterWorkflowRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendWorkflowRun(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/SendWorkflowRun',
            requests__pb2.SendWorkflowRunRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetWorkflowRunArgs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetWorkflowRunArgs',
            requests__pb2.GetWorkflowRunArgsRequest.SerializeToString,
            requests__pb2.GetWorkflowRunArgsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AdvanceWorkflowRun(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/AdvanceWorkflowRun',
            requests__pb2.AdvanceWorkflowRunRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetWorkflowDurableRuns(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetWorkflowDurableRuns',
            requests__pb2.GetWorkflowDurableRunsRequest.SerializeToString,
            requests__pb2.GetWorkflowDurableRunsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



import logging
import os
import signal
from datetime import datetime, timezone
from multiprocessing import Event, Process

from croniter import croniter

from hyrex.dispatcher import CronJob, get_dispatcher
from hyrex.dispatcher.dispatcher import CronJobRun
from hyrex.worker.logging import LogLevel, init_logging
from hyrex.worker.utils import is_process_alive

DEFAULT_HEARTBEAT_INTERVAL_SECONDS = 10
LOOP_RATE_SECONDS = 30


class WorkerCronScheduler(Process):
    def __init__(self, log_level: LogLevel, worker_name: str):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self.log_level = log_level
        self.worker_name = worker_name

        self._stop_event = Event()

        # To check if root process is running
        self.parent_pid = os.getpid()

    def check_stop_conditions(self):
        # Confirm parent is still alive
        if not is_process_alive(self.parent_pid):
            self.logger.warning("Root process died unexpectedly. Shutting down.")
            self._stop_event.set()

    def acquire_scheduler_lock(self) -> int | None:
        self.logger.info("Acquiring cron scheduler lock...")
        result = self.dispatcher.acquire_scheduler_lock(self.worker_name)
        return result

    def update_cron_confirmation_timestamp_to_now(self, cron_job_id: int):
        self.dispatcher.update_cron_job_confirmation_timestamp(cron_job_id)

    def impute_scheduled_cron_job_runs(self, cron_job: CronJob) -> list[CronJobRun]:
        # Create iterator starting from the last confirmed date or activation date
        start_time = max(cron_job.activated_at, cron_job.scheduled_jobs_confirmed_until)
        iterator = croniter(cron_job.schedule, start_time=start_time)

        cron_job_runs = []
        now = datetime.now(timezone.utc)  # Create timezone-aware UTC datetime

        next_interval_date = iterator.get_next(datetime)
        # Ensure next_interval_date is timezone-aware with UTC
        if next_interval_date.tzinfo is None:
            next_interval_date = next_interval_date.replace(tzinfo=timezone.utc)

        while next_interval_date <= now:
            cron_job_runs.append(
                CronJobRun(
                    jobid=cron_job.jobid,
                    command=cron_job.command,
                    schedule_time=next_interval_date,
                )
            )

            next_interval_date = iterator.get_next(datetime)
            if next_interval_date.tzinfo is None:
                next_interval_date = next_interval_date.replace(tzinfo=timezone.utc)

        return cron_job_runs

    def run(self):
        init_logging(self.log_level)

        self.logger.info("Initializing cron scheduler.")
        self.dispatcher = get_dispatcher()

        # Ignore signals, let main process manage shutdown.
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        self.lock_id = None

        try:
            while not self._stop_event.is_set() and not self.lock_id:
                self.check_stop_conditions()

                self.lock_id = self.acquire_scheduler_lock()
                if not self.lock_id and not self._stop_event.is_set():
                    self.logger.info(
                        "Failed to acquire lock. Trying again in 15 seconds."
                    )
                    # Sleep and then try again
                    self._stop_event.wait(15)
                    continue

            # If we're stopping, exit early
            if self._stop_event.is_set():
                self.logger.info("Stop event detected during lock acquisition.")
                return

            self.logger.info("Acquired lock.")

            # Decide whether to backfill cron jobs
            cron_expressions = self.dispatcher.pull_cron_job_expressions()
            for cron_job in cron_expressions:
                if not cron_job.should_backfill:
                    self.update_cron_confirmation_timestamp_to_now(cron_job.jobid)

            # Main loop with lock held
            while not self._stop_event.is_set():
                self.check_stop_conditions()

                # TODO: Update lock heartbeat

                cron_expressions = self.dispatcher.pull_cron_job_expressions()

                # Queue cron job runs
                for cron_job in cron_expressions:
                    self.logger.info(
                        f"Got cron job {cron_job.jobname}, confirmed_until={cron_job.scheduled_jobs_confirmed_until}"
                    )
                    scheduled_jobs = self.impute_scheduled_cron_job_runs(cron_job)
                    self.dispatcher.schedule_cron_job_runs(scheduled_jobs)

                # Execute cron job runs
                result = self.dispatcher.execute_queued_cron_job_run()
                while result and result == "executed":
                    self.logger.info("Executed cron job run...")
                    result = self.dispatcher.execute_queued_cron_job_run()

                self._stop_event.wait(LOOP_RATE_SECONDS)
        finally:
            self.stop()

    def stop(self):
        self.logger.info("Stopping cron scheduler.")
        if self.lock_id:
            self.logger.info("Releasing scheduler lock...")
            self.dispatcher.release_scheduler_lock(self.worker_name)
            self.lock_id = None
        self.dispatcher.stop()
